#!/usr/bin/env python
"""A library for building/decomposing URIs
"""

__author__     = 'Brandon Sandrowicz'
__version__    = '0.2'
__copyright__  = 'Copyright 2012, Brandon Sandrowicz'
__license__    = 'MIT'
__maintainer__ = 'Brandon Sandrowicz'
__email__      = 'brandon@sandrowicz.org'
__status__     = 'Development'

import re

# Character Classes
sub_delims  = "[!$&'()*+,;=]"
pct_encoded = "%[a-fA-F0-9]{2}"
unreserved  = "[a-zA-Z0-9._~-]"

# Parts
scheme_re    = re.compile(r'[^\W0-9_]([^\W_]|[+.-])*')
hier_part_re = re.compile(r'(%s|%s|%s|/)*' % (sub_delims, pct_encoded, unreserved))

def join_if_exists(join_string, iterable):
    return join_string.join(item for item in iterable if item)

class URI(object):
    orig_uri  = ""
    scheme    = ""
    hier_part = ""
    query     = ""
    fragment  = ""
    authority = ""
    path      = ""

    def __init__(self, uri):
        self.orig_uri = uri.strip()
        self._process_uri()

    def __str__(self):
        return self.uri

    def __repr__(self):
        return "<URI('%s')>" % self.uri

    def _process_uri(self):
        uri = self.orig_uri

        # Scheme
        parts = uri.split(':', 1)
        if len(parts) == 2:
            scheme, rest = parts
            match = scheme_re.match(scheme)
            if match:
                self.scheme = scheme
                uri = rest

        # Fragment
        parts = uri.rsplit('#', 1)
        if len(parts) == 2:
            rest, fragment = parts
            self.fragment = fragment
            uri = rest

        # Hier Part
        parts = uri.split('?', 1)
        if len(parts) == 2:
            hier_part, query = parts
        else:
            hier_part = uri
            query = ""

        match = hier_part_re.match(hier_part)
        if match:
            self.hier_part = hier_part
        else:
            self.hier_part = None

        self.query = query

        self._process_hier_part()

    def _process_hier_part(self):
        if self.hier_part.startswith('//'):
            parts = (self.hier_part[2:]).split('/', 1)
            if len(parts) == 2:
                self.authority, self.path = parts
                self.path = '/' + self.path
            else:
                self.authority = self.hier_part[2:]
        else:
            self.path = self.hier_part

    @property
    def uri(self):
        uri = join_if_exists(':', [self.scheme, self.hier_part])
        uri = join_if_exists('?', [uri, self.query])
        uri = join_if_exists('#', [uri, self.fragment])
        return uri

    @classmethod
    def parse_uri(cls, uri):
        uri = cls(uri)
        return {
            'scheme'  : uri.scheme,
            'netloc'  : uri.authority,
            'path'    : uri.path,
            'params'  : uri.query,
            'fragment': uri.fragment,
        }




"""
Start the basic thor servers for each case of connection.
Use the standard request responders in BaseServer.
"""
import yaml
import logging
import argparse
from thor.loop import run
from thor import HttpServer
from thor import SpdyServer
from serverbase import BaseServer
from multiprocessing import Process
from thor.events import on
import urilib
import util
import os


def get_args():
    parser = argparse.ArgumentParser(
            description='Run SPDY and HTTP servers to act as endpoint of test.',
            epilog='This can be run directly or using `scripts/run-servers`.')
    parser.add_argument('-c', '-config', type=str, required=False,
            dest='conf_file', action='store', help='YAML configuration file.')
    return parser.parse_args()


def http_main(conf):
    print 'Loading configuration into http base server!'
    base = BaseServer(conf)
    print 'creating HTTP server on port %d' % (conf['http_port'],)

    def http_handler(x):
        @on(x, 'request_start')
        def go(*args):
            print 'HTTP: start %s' % (str(args[1]),)
            base.paths.get(util.make_ident(x.method, x.uri), base.fourohfour)(x)

        @on(x, 'request_body')
        def body(chunk):
            #print 'body: %s' % chunk
            pass

        @on(x, 'request_done')
        def done(trailers):
            #print 'done: %s' % str(trailers)
            pass

    http_serve = HttpServer(host='', port=conf['http_port'])
    http_serve.on('exchange', http_handler)


def spdy_main(conf):
    print 'Loading configuration into spdy base server!'
    base = BaseServer(conf)
    print 'creating SPDY server on port %d' % (conf['spdy_port'],)

    def spdy_handler(x):
        @on(x, 'request_start')
        def go(*args):
            path = urilib.URI(x.uri).path  # Take into account SPDY's uri
            print 'SPDY: start %s' % (path,)
            base.paths.get(util.make_ident(x.method, path), base.fourohfour)(x)

        @on(x, 'request_body')
        def body(chunk):
            #print 'body: %s' % chunk
            pass

        @on(x, 'request_done')
        def done(trailers):
            #print 'done: %s' % str(trailers)
            pass

    spdy_serve = SpdyServer('', conf['spdy_port'])
    spdy_serve.on('exchange', spdy_handler)


def capture_main(conf):
    print 'creating capture server on port %d' % (conf['capture_port'],)

    def capture_handler(x):
        up = {'inp': '', 'final': {}}

        @on(x, 'request_start')
        def go(*args):
            print 'capture: start %s' % (str(args[1]),)

        @on(x, 'request_body')
        def body(chunk):
            up['inp'] += chunk

        @on(x, 'request_done')
        def done(trailers):
            with open(os.path.join(conf['outdir'], conf['outfile']), 'a') as f:
                f.write(up['inp'])
                f.write('\n')
                f.close()
            print 'finished capture'
            x.response_start(200, 'OK', [])
            x.response_done([])

    capture_serve = HttpServer(host='', port=conf['capture_port'])
    capture_serve.on('exchange', capture_handler)

if __name__ == '__main__':
    args = get_args()
    #conf = yaml.load(file(args.conf_file, 'r'))
    dirpath = os.path.dirname(__file__)
    conf = yaml.load(file(os.path.join(dirpath, 'conf.yaml'), 'r'))

    try:
        os.mkdir(conf['outdir'])
    except OSError:
        pass

    capture_main(conf)
    spdy_main(conf)
    http_main(conf)
    run()



"""
Base class for common operations of servers.
"""
import util
import copy
import json


class BaseServer(object):

    def __init__(self, conf):
        self.paths = {}
        util.register_all(self)
        self.sitelist = util.load_list(conf['sitelist'])
        self.sites = {}
        for i,line in enumerate(open(util.rel_path(conf['sitedump']), 'r')):
            if str(i) in self.sitelist:
                self.sites[str(i)] = json.loads(line)

    # HTTP verbs and paths below

    def fourohfour(self, res):
        headers = []
        res.response_start(404, 'Error Not Found', headers)
        res.response_body('THAT DOESN\'T EXIST!')
        res.response_done([])

    """
    Send the list of sites to access.
    """
    @util.register(method='GET', path='/hello')
    def gethello(self, res):
        headers = []
        res.response_start(200, 'OK', headers)
        res.response_body(json.dumps(self.sitelist))
        res.response_done([])

    """
    For a given site, return the assets to acquire.
    """
    @util.register(method='GET', path='/site/')
    def getsite(self, res):
        headers = []
        index = int(res.uri.split('/')[-1])
        data = {'index': index}
        try:
            site = copy.deepcopy(self.sites[str(index)])
        except KeyError:
            self.fourohfour(res)
            return
        if int(site['assets']['asset0']['size']) == 0:
            reqsize = int(site['assets'].pop('asset1', {'size':0})['size'])
            site['assets'].pop('asset0')
        else:
            reqsize = int(site['assets'].pop('asset0')['size'])
        data['list'] = util.make_assets(site)
        data['junk'] = util.fill_junk(reqsize - len(str(data)))
        res.response_start(200, 'OK', headers)
        res.response_body(json.dumps(data))
        res.response_done([])

    """
    For a given asset, return a blob that is the size of the asset.
    Take into account compressibility.
    """
    @util.register(method='GET', path='/asset/')
    def getasset(self, res):
        headers = []
        asset = res.uri.split('/')[-1]
        size, comp = asset.split('.')
        size = int(size)
        # FIXME: This is not necessarily what the compression factor should be
        if comp == 'c':
            size *= 0.7
        junk = util.fill_junk(int(size))
        res.response_start(200, 'OK', headers)
        res.response_body(junk)
        res.response_done([])



"""
Client to run the tests
"""
import json
import os
import time
import yaml
import argparse
from progress.bar import Bar
from thor import HttpClient
from thor import SpdyClient
from thor.loop import stop, run


def get_args():
    parser = argparse.ArgumentParser(
            description='Run HTTP and SPDY clients for test.')
    parser.add_argument('-c', '-config', type=str, required=False,
            dest='conf_file', action='store', help='YAML configuration file.')
    return parser.parse_args()


def hello(host):
    resp = {'text': '', 'sitelist': []}

    def hello_start(status, phrase, headers):
        if status == '200':
            print 'Connection Successful!\nDownloading sitelist.'

    def hello_body(chunk):
        resp['text'] += chunk

    def hello_stop(trailers):
        resp['sitelist'] = json.loads(resp['text'])
        stop()

    def hello_err(err):
        print 'Connection could not be completed...'
        stop()
        exit()

    httpclient = HttpClient()
    httpclient.connect_timeout = 5
    hello = httpclient.exchange()
    hello.on('response_start', hello_start)
    hello.on('response_body', hello_body)
    hello.on('response_done', hello_stop)
    hello.on('error', hello_err)
    uri = host + '/hello'
    hello.request_start('GET', uri, [])
    hello.request_done([])
    run()
    return resp['sitelist']


def collect(host, data):

    def collect_start(status, phrase, headers):
        if status != '200':
            print 'Uploading data...'

    def collect_body(chunk):
        pass
        #print chunk

    def collect_stop(trailers):
        #print 'Finished!'
        stop()

    httpclient = HttpClient()
    collect = httpclient.exchange()
    collect.on('response_start', collect_start)
    collect.on('response_body', collect_body)
    collect.on('response_done', collect_stop)
    collect.request_start('POST', host, [])
    collect.request_body(data)
    collect.request_done([])
    run()
    return "Upload Complete"


def http_siteget(http1, http2, site):
    resp = {'text': '', 'assetlist': []}

    def site_start(status, phrase, headers):
        if status != '200':
            print 'Connection Failed: ' + site

    def site_body(chunk):
        resp['text'] += chunk

    def site_stop(trailers):
        resp['assetlist'] = json.loads(resp['text'])['list']
        stop()

    t0 = time.time()
    httpclient = HttpClient()
    get = httpclient.exchange()
    get.on('response_start', site_start)
    get.on('response_body', site_body)
    get.on('response_done', site_stop)
    uri = http1 + '/site/' + str(site)
    get.request_start('GET', uri, [])
    get.request_done([])
    run()
    for asset in resp['assetlist']:
        # TODO: consider doing this with 4-6 connections?
        http_assetget(http1, http2, asset, httpclient)
    return time.time() - t0


def http_assetget(http1, http2, asset, httpclient):
    resp = {'text': ''}

    def asset_start(status, phrase, headers):
        if status != '200':
            print 'Connection Failed: ' + asset
            print status
            print phrase

    def asset_body(chunk):
        resp['text'] += chunk

    def asset_stop(trailers):
        stop()

    get = httpclient.exchange()
    get.on('response_start', asset_start)
    get.on('response_body', asset_body)
    get.on('response_done', asset_stop)
    parts = asset.split('/')
    if parts[0] == 'host1':
        uri = http1 + '/asset/' + parts[1]
    else:
        uri = http2 + '/asset/' + parts[1]
    get.request_start('GET', uri, [])
    get.request_done([])
    run()


def spdy_siteget(spdy1, spdy2, site):
    resp = {'text': '', 'assetlist': []}

    def site_start(status, phrase, headers):
        if status != '200':
            print 'Connection Failed: ' + site

    def site_body(chunk):
        resp['text'] += chunk

    def site_stop(trailers):
        resp['assetlist'] = json.loads(resp['text'])['list']
        stop()

    t0 = time.time()
    spdyclient = SpdyClient()
    get = spdyclient.exchange()
    uri = spdy1 + '/site/' + str(site)
    stream = get.request_start('GET', uri, [])
    get._streams[stream].on('response_start', site_start)
    get._streams[stream].on('response_body', site_body)
    get._streams[stream].on('response_done', site_stop)
    get._streams[stream].request_done(stream, [])
    run()
    for asset in resp['assetlist']:
        spdy_assetget(spdy1, spdy2, asset, spdyclient)
    return time.time() - t0


def spdy_assetget(spdy1, spdy2, asset, spdyclient):
    resp = {'text': ''}

    def asset_start(status, phrase, headers):
        if status != '200':
            print 'Connection Failed: ' + asset

    def asset_body(chunk):
        resp['text'] += chunk

    def asset_stop(trailers):
        stop()

    get = spdyclient.exchange()
    parts = asset.split('/')
    if parts[0] == 'host1':
        uri = str(spdy1 + '/asset/' + parts[1])
    else:
        uri = str(spdy2 + '/asset/' + parts[1])
    stream = get.request_start('GET', uri, [])
    get._streams[stream].on('response_start', asset_start)
    get._streams[stream].on('response_body', asset_body)
    get._streams[stream].on('response_done', asset_stop)
    get._streams[stream].request_done(stream, [])
    run()


def main():
    args = get_args()
    #conf = yaml.load(file(args.conf_file, 'r'))
    dirpath = os.path.dirname(__file__)
    conf = yaml.load(file(os.path.join(dirpath, 'conf.yaml'), 'r'))

    mainhost_http = conf['mainhost'] + ':' + str(conf['http_port'])
    mainhost_spdy = conf['mainhost'] + ':' + str(conf['spdy_port'])
    secondhost_http = conf['secondhost'] + ':' + str(conf['http_port'])
    secondhost_spdy = conf['secondhost'] + ':' + str(conf['spdy_port'])
    mainhost_collect = conf['mainhost'] + ':' + str(conf['capture_port'])

    times = {}
    sites = hello(mainhost_http)
    bar = Bar('Running Test', max=len(sites))
    for i,site in enumerate(sites):
        bar.next()
        # TODO: MAKE THIS PRINT HELPFUL STUFF AND JUST DO IT ON ONE LINE!
        http_delta = http_siteget(mainhost_http,
                secondhost_http,
                site)
        spdy_delta = spdy_siteget(mainhost_spdy,
                secondhost_spdy,
                site)
        times[site] = {}
        times[site]['http'] = http_delta
        times[site]['spdy'] = spdy_delta

    bar.finish()
    print 'Testing complete!'
    result = collect(mainhost_collect, json.dumps(times))
    print result
    stop()


if __name__ == '__main__':
    main()



"""
General purpose helpers and decorators
"""
import random
import os
import string


def register_all(cls):
    for methodname in dir(cls):
        method = getattr(cls, methodname)
        if hasattr(method, '_route') and hasattr(method, '_method'):
            cls.paths[make_ident(method._method, method._route)] = method


def register(method, path, *args):
    def wrapper(func):
        func._route = path
        func._method = method
        return func
    return wrapper

def rel_path(f):
    dirpath = os.path.dirname(__file__)
    return os.path.join(dirpath, f)

def make_ident(method, route):
    """
    For the time being, if a path ends in '/', it expects an argument to follow
    """
    # FIXME: This path stuff is dumb. Do this better if time permits
    parts = route.rstrip('/1234567890.cu')
    return method + '->' + parts


def make_ident_spdy(method, route):
    """
    For the time being, if a path ends in '/', it expects an argument to follow
    """
    # FIXME: This path stuff is dumb. Do this better if time permits
    parts = route.rstrip('/1234567890.cu').split('/')[-1]
    return method + '->' + parts


def fill_junk(size):
    return ''.join([random.choice(string.letters) for i in xrange(size)])


def load_list(f):
    ls = []
    for line in open(rel_path(f), 'r'):
        ls.append(line.split()[0])
    return ls


def make_assets(site):
    assets = []
    for asset in site['assets'].values():
        if asset['compressible']:
            trailer = '.c'
        else:
            trailer = '.u'
        if asset['remote']:
            header = 'host2/'
        else:
            header = 'host1/'
        assets.append(header + str(asset['size']) + trailer)
    return assets


class resmap(object):

    def __init__(self, func, uri):
        self.func = func
        self.uri = uri

    def response_start(self, code, phrase, headers):
        x, y = self.func(code,
                phrase,
                headers,
                dummy)
        self.response_body = self.resp_body_wrap(x)
        self.response_done = self.resp_done_wrap(y)

    def resp_body_wrap(self, func):
        def wrap(body):
            func(body)
        return wrap

    def resp_done_wrap(self, func):
        def wrap(body):
            func(body)
        return wrap
